// Fichier libre de droits

#ifndef IMPLEMENTATION32BITS_HPP
#define IMPLEMENTATION32BITS_HPP

#include <limits>
#include "ImplementationBit.h"
#include "Exceptions.h"
#include "Mouvement.h"


template<class integer>
ImplementationBit<integer>::Point::Point(int lig, int col) {
	Assert(col >= 0);
	Assert(col < getMaximumSize());
	descLigne = 1;
	descLigne <<= col;
	numeroLigne = lig;
}

template<class integer>
void ImplementationBit<integer>::Point::display(std::ostream &os) const {
	int colonne = 0;
	integer mask = descLigne;
	while (mask != 0) {
		if (1 & mask) {
			os << "(" << numeroLigne << ", " << colonne << ")";
		}
		colonne += 1;
		mask >>= 1;
	}
}

template<class integer>
bool ImplementationBit<integer>::operator==(const ImplementationBit<integer> &d) const {
	Assert(isHexagonal == d.isHexagonal);
	Assert(height == d.height);
	for (int ligne = 0; ligne < height; ++ligne) {
		if (lignes[ligne] != d.lignes[ligne]) {
			return false;
		}
	}
	return true;
}

template<class integer>
bool ImplementationBit<integer>::operator<(const ImplementationBit<integer> &d) const {
	Assert(isHexagonal == d.isHexagonal);
	Assert(height == d.height);
	for (int ligne = 0; ligne < height; ++ligne) {
		if (lignes[ligne] != d.lignes[ligne]) {
			return lignes[ligne] < d.lignes[ligne];
		}
	}
	return false;
}

template<class integer>
void ImplementationBit<integer>::init(bool hexa, int indiceLigneMax) {
	isHexagonal = hexa;
	delete lignes;
	lignes = 0;
	height = indiceLigneMax + 1;
	if (height > 0) {
		lignes = new integer [height];
		for(int ligne = 0; ligne < height; ++ligne) {
			lignes[ligne] = 0;
		}
	} else {
		lignes = 0;
	}
}

template<class integer>
ImplementationBit<integer>::ImplementationBit() {
	isHexagonal = false;
	lignes = 0;
	height = 0;
}

template<class integer>
ImplementationBit<integer>::~ImplementationBit() {
	delete [] lignes;
}

template<class integer>
bool ImplementationBit<integer>::libre(const Point &p) const {
	Assert(p.numeroLigne <= height);
	return (0 == (p.descLigne & lignes[p.numeroLigne]));
}

template<class integer>
void ImplementationBit<integer>::ajoute(const Point &p) {
	Assert(p.numeroLigne <= height);
	lignes[p.numeroLigne] |= p.descLigne;
}

template<class integer>
void ImplementationBit<integer>::retire(const Point &p) {
	Assert(p.numeroLigne <= height);
	lignes[p.numeroLigne] &= ~p.descLigne;
}

template<class integer>
int ImplementationBit<integer>::taille(void) const {
	int result = 0;
	for (int ligne = 0; ligne < height; ++ligne) {
		for (integer mask = lignes[ligne]; mask != 0; mask >>= 1) {
			if (1 & mask) {
				result += 1;
			}
		}
	}
	return result;
}

template<class integer>
void ImplementationBit<integer>::display(std::ostream &os, const char *caractereOccupe) const {
	std::string caseLibre(" ");
	std::string caseOccupee(caractereOccupe);
	if (isHexagonal) {
		caseLibre += " ";
		caseOccupee += " ";
	}
	for (int ligne = 0; ligne < height; ++ligne) {
		if (isHexagonal) {
			os << std::string(ligne, ' ');
		}
		for (integer mask = lignes[ligne]; mask != 0; mask >>= 1) {
			if (1 & mask) {
				os << caseOccupee;
			} else {
				os << caseLibre;
			}
		}
		os << '\n';
	}
}

template<class integer>
void ImplementationBit<integer>::initListeMouvements(void) {
	listeMouvements.clear();
	for (int ligne = 0; ligne < height; ++ligne) {
		int colonne = 0;
		for (integer mask = lignes[ligne]; mask != 0; mask >>= 1) {
			if (7 == (7 & mask)) {
				listeMouvements.push_back(Mouvement<Point>(Point(ligne, colonne), Point(ligne, colonne + 1), Point(ligne, colonne + 2)));
				listeMouvements.push_back(Mouvement<Point>(Point(ligne, colonne + 2), Point(ligne, colonne + 1), Point(ligne, colonne)));
			}
			colonne += 1;
		}
	}
	for (int ligne = 0; ligne < height - 2; ++ligne) {
		int colonne = 0;
		for (integer mask = lignes[ligne] & lignes[ligne + 1] & lignes[ligne + 2]; mask != 0; mask >>= 1) {
			if (1 & mask) {
				listeMouvements.push_back(Mouvement<Point>(Point(ligne, colonne), Point(ligne + 1, colonne), Point(ligne + 2, colonne)));
				listeMouvements.push_back(Mouvement<Point>(Point(ligne + 2, colonne), Point(ligne + 1, colonne), Point(ligne, colonne)));
			}
			colonne += 1;
		}
	}
	if (isHexagonal) {
		for (int ligne = 0; ligne < height - 2; ++ligne) {
			int colonne = 0;
			for (integer mask = (lignes[ligne] >> 2) & (lignes[ligne + 1] >> 1) & lignes[ligne + 2]; mask != 0; mask >>= 1) {
				if (1 & mask) {
					listeMouvements.push_back(Mouvement<Point>(Point(ligne, colonne + 2), Point(ligne + 1, colonne + 1), Point(ligne + 2, colonne)));
					listeMouvements.push_back(Mouvement<Point>(Point(ligne + 2, colonne), Point(ligne + 1, colonne + 1), Point(ligne, colonne + 2)));
				}
				colonne += 1;
			}
		}
	}
}

template<class integer>
void ImplementationBit<integer>::swap(ImplementationBit &e) {
	bool tempBool = e.isHexagonal;
	e.isHexagonal = isHexagonal;
	isHexagonal = tempBool;
	int tempInt = e.height;
	e.height = height;
	height = tempInt;
	integer *tempPtr = e.lignes;
	e.lignes = lignes;
	lignes = tempPtr;
}

template<class integer>
ImplementationBit<integer> &ImplementationBit<integer>::operator=(const ImplementationBit<integer> &e) {
	if (this != &e) {
		ImplementationBit<integer> temp(e);
		swap(temp);
	}
	return *this;
}

template<class integer>
ImplementationBit<integer>::ImplementationBit(const ImplementationBit<integer> &e) {
	height = e.height;
	lignes = 0;
	if (height > 0) {
		lignes = new integer [height];
		for (int ligne = 0; ligne < height; ++ligne) {
			lignes[ligne] = e.lignes[ligne];
		}
	}
	isHexagonal = e.isHexagonal;
}

template<class integer>
int ImplementationBit<integer>::getMaximumSize(void) {
	return std::numeric_limits<integer>::digits;
}

#endif // IMPLEMENTATION32BITS_HPP
